@@ -8,6 +8,7 @@
 #' @param ... see [kronecker()]
 #' @return if 'X' and 'Y' are `cgeneric`
 #' return a `cgeneric`, else a `rgeneric`.
+#' @importFrom inlabru bru_get_mapper
 NULL
 #> NULL
 
@@ -43,25 +44,18 @@ setMethod(
     } else {
       useINLAprecomp = mcall$useINLAprecomp
     }
-    if(is.null(mcall$libpath)) {
-      libpath <- NULL
-    } else {
-      libpath <- mcall$libpath
-    }
 
-    model <- "inla_cgeneric_kronecker"
-    if (is.null(libpath)) {
-      if (useINLAprecomp) {
-        libpath <- INLA::inla.external.lib("INLAtools")
-      } else {
-        libpath <- system.file("libs", package = "INLAtools")
-        if (Sys.info()["sysname"] == "Windows") {
-          libpath <- file.path(libpath, "x64/INLAtools.dll")
-        } else {
-          libpath <- file.path(libpath, "INLAtools.so")
-        }
-      }
+    INLAvcheck <- packageCheck("INLA", "25-10-28")
+    if(is.na(INLAvcheck) & useINLAprecomp) {
+      useINLAprecomp <- FALSE
+      warning("INLA version is old. Setting 'useINLAprecomp = FALSE'!")
     }
+    shlib <- cgeneric_shlib(
+      package = "INLAtools",
+      useINLAprecomp = useINLAprecomp,
+      debug = debug)
+
+    cmodel <- "inla_cgeneric_kronecker"
 
     n1 <- as.integer(X$f$n)
     n2 <- as.integer(Y$f$n)
@@ -166,19 +160,29 @@ setMethod(
     )
 
     ## initial data
-    ret <- list(
-      f = list(
-        model = "cgeneric",
-        n = as.integer(N),
-        cgeneric = list(
-          model = model,
-          shlib = libpath,
+    ret <- structure(
+      list(
+        f = list(
+          model = "cgeneric",
           n = as.integer(N),
-          debug = as.integer(debug),
-          data = list()
+          cgeneric = structure(
+            list(
+              model = cmodel,
+              shlib = shlib,
+              n = as.integer(N),
+              debug = as.integer(debug)
+            ),
+            # inla.cgeneric is needed to support INLA before August 2025
+            class = c("inla.cgeneric.f", "inla.cgeneric")
+          )
         )
-      )
+      ),
+      class = c("cgeneric", "inla.cgeneric")
     )
+    ret$f$cgeneric$data <- vector("list", 5L)
+    names(ret$f$cgeneric$data) <- c(
+      "ints", "doubles", "characters",
+      "matrices", "smatrices")
 
     ## data size for each model
     ndata1 <- sapply(
@@ -210,27 +214,31 @@ setMethod(
           )
         )
 
-    ret$f$cgeneric$data$doubles <-
-      c(
-        X$f$cgeneric$data$doubles,
-        Y$f$cgeneric$data$doubles
-      )
+      if((ndata1[2]>0) | (ndata2[2]>0)) {
+        ret$f$cgeneric$data$doubles <-
+          c(
+            X$f$cgeneric$data$doubles,
+            Y$f$cgeneric$data$doubles
+          )
+      }
 
-    ret$f$cgeneric$data$characters <-
+      ret$f$cgeneric$data$characters <-
       c(
         list(
-          model = model,
-          shlib = libpath
+          model = cmodel,
+          shlib = shlib
         ),
         X$f$cgeneric$data$characters,
         Y$f$cgeneric$data$characters
       )
 
-    ret$f$cgeneric$data$matrices <-
-      c(
-        X$f$cgeneric$data$matrices,
-        Y$f$cgeneric$data$matrices
-      )
+      if((ndata1[4]>0) | (ndata2[4]>0)) {
+        ret$f$cgeneric$data$matrices <-
+          c(
+            X$f$cgeneric$data$matrices,
+            Y$f$cgeneric$data$matrices
+          )
+      }
 
     ret$f$cgeneric$data$smatrices <-
       c(
@@ -245,42 +253,37 @@ setMethod(
         )
       )
 
-    class(ret) <- "cgeneric"
-    class(ret$f$cgeneric) <- "cgeneric"
+    if((!is.null(X$f$extraconstr)) |
+       (!is.null(Y$f$extraconstr))) {
+        ret$f$extraconstr <-
+          kronecker_extraconstr(
+            X$f$extraconstr,
+            Y$f$extraconstr,
+            X$f$n, Y$f$n
+          )
+        if(debug) {
+          cat(nrow(ret$f$extraconstr),
+              " 'extraconstr' built!\n")
+        }
+    }
 
-    if(is.null(X$f$extraconstr)) {
-      if(is.null(Y$f$extraconstr)) {
-        if(debug)
-          cat("No extraconstr!\n")
-      } else {
-        c2 <- Y$f$extraconstr
-        ret$f$extraconstr <- list(
-          A = kronecker(diag(X$f$n), c2$A),
-          e = rep(c2$e, X$f$n)
-        )
-      }
+    # Note: kron(X,Y) gives X-major ordering (the X-index varies slowly, the
+    # Y-index varies quickly), which requires the mappers to be in reverse
+    # order, multi(Y,X):
+    inlabruCheck <- packageCheck("inlabru", "2.13.0.9005")
+    if(is.na(inlabruCheck)) {
+      warning("Please install a inlabru recent version.")
     } else {
-      c1 <- X$f$extraconstr
-      if(is.null(Y$f$extraconstr)) {
-        ret$f$extraconstr <- list(
-          A = kronecker(c1$A, diag(Y$f$n)),
-          e = rep(c1$e, each = Y$f$n)
+      ret$mapper <-
+      inlabru::bm_multi(
+        list(
+          inlabru::bru_get_mapper(Y),
+          inlabru::bru_get_mapper(X)
         )
-      } else {
-        c2 <- Y$f$extraconstr
-        ret$f$extraconstr <- list(
-          A = rbind(
-            kronecker(c1$A, diag(ncol(c2$A))),
-            kronecker(diag(ncol(c1$A)), c2$A)
-          ),
-          e = c(rep(c1$e, each = ncol(c2$A)),
-                rep(c2$e, ncol(c1$A)))
-        )
-      }
+      )
     }
 
     return(ret)
-
   }
 )
 #' @rdname kronecker
@@ -315,15 +318,15 @@ setMethod(
                        theta = NULL) {
 
       graph <- function(n, theta) {
-        g1 <- cgeneric_get(X, "graph", optimize = FALSE)
-        g2 <- INLA::inla.rgeneric.q(Y, "graph")
-        return(kronecker(g1, g2))
+        g1 <- cgeneric_get(X, cmd = "graph", optimize = FALSE)
+        g2 <- rgeneric_get(Y, cmd = "graph", optimize = FALSE)
+        return(Sparse(kronecker(g1, g2)))
       }
 
       Q <- function(n, theta) {
         Q1 <- cgeneric_get(X, "Q", theta = theta[1:nth1], optimize = FALSE)
-        Q2 <- INLA::inla.rgeneric.q(Y, "Q", theta = theta[nth1+1:nth2], optimize = FALSE)
-        QQ <- INLA::inla.as.sparse(kronecker(Q1, Q2))
+        Q2 <- rgeneric_get(Y, "Q", theta = theta[nth1+1:nth2], optimize = FALSE)
+        QQ <- Sparse(kronecker(Q1, Q2))
         idx <- which(QQ@i <= QQ@j)
         return(QQ@x[idx])
       }
@@ -336,8 +339,8 @@ setMethod(
 
       log.prior <- function(n, theta) {
         return(
-          prior(X, theta = theta[1:nth1]) +
-            prior(Y, theta = theta[nth1+1:nth2])
+          cgeneric_get(X, cmd = "log_prior", theta = theta[1:nth1]) +
+            rgeneric_get(Y, cmd = "log_prior", theta = theta[nth1+1:nth2])
         )
       }
 
@@ -365,24 +368,59 @@ setMethod(
     }
 
     ### follows INLA:::inla.rgeneric.define() but no assign env
-    rmodel <- list(
-      f = list(
-        model = "rgeneric",
-        n = n,
-        rgeneric = list(
-          definition =
-            compiler::cmpfun(
-              kmodel,
-              options = list(optimize = 3L)),
-          debug = debug,
-          optimize = TRUE
+    rmodel <- structure(
+      list(
+        f = list(
+          model = "rgeneric",
+          n = n,
+          rgeneric = structure(
+            list(
+              definition =
+                compiler::cmpfun(
+                  kmodel,
+                  options = list(optimize = 3L)),
+              debug = debug,
+              optimize = TRUE
+            ),
+            # inla.rgeneric is needed to support INLA before August 2025
+            class = c("inla.rgeneric.f", "inla.rgeneric")
+          )
+        )
+      ),
+      class = c("rgeneric", "inla.rgeneric")
+    )
+
+    if((!is.null(X$f$extraconstr)) |
+       (!is.null(Y$f$extraconstr))) {
+      rmodel$f$extraconstr <-
+        kronecker_extraconstr(
+          X$f$extraconstr,
+          Y$f$extraconstr,
+          X$f$n, Y$f$n
+        )
+      if(debug) {
+        cat(nrow(rmodel$f$extraconstr),
+            " 'extraconstr' built\n")
+      }
+    }
+
+    # Note: kron(X,Y) gives X-major ordering (the X-index varies slowly, the
+    # Y-index varies quickly), which requires the mappers to be in reverse
+    # order, multi(Y,X):
+    inlabruCheck <- packageCheck("inlabru", "2.13.0.9005")
+    if(is.na(inlabruCheck)) {
+      warning("Please install a inlabru recent version.")
+    } else {
+      rmodel$mapper <-
+      inlabru::bm_multi(
+        list(
+          inlabru::bru_get_mapper(Y),
+          inlabru::bru_get_mapper(X)
         )
       )
-    )
-    class(rmodel) <- "rgeneric"
-    class(rmodel$f$rgeneric) <- "rgeneric"
-    return(rmodel)
+    }
 
+    return(rmodel)
   }
 )
 #' @rdname kronecker
@@ -417,15 +455,15 @@ setMethod(
                        theta = NULL) {
 
       graph <- function(n, theta) {
-        g1 <- INLA::inla.rgeneric.q(rmodel = X, cmd = "graph")
-        g2 <- cgeneric_get(Y, "graph", optimize = FALSE)
-        return(kronecker(g1, g2))
+        g1 <- rgeneric_get(X, cmd = "graph", optimize = FALSE)
+        g2 <- cgeneric_get(Y, cmd = "graph", optimize = FALSE)
+        return(Sparse(kronecker(g1, g2)))
       }
 
       Q <- function(n, theta) {
-        Q1 <- INLA::inla.rgeneric.q(rmodel = X, cmd = "Q", theta = theta[1:nth1])
-        Q2 <- cgeneric_get(Y, "Q", theta = theta[nth1+1:nth2], optimize = FALSE)
-        QQ <- INLA::inla.as.sparse(kronecker(Q1, Q2))
+        Q1 <- rgeneric_get(X, cmd = "Q", theta = theta[1:nth1], optimize = FALSE)
+        Q2 <- cgeneric_get(Y, cmd = "Q", theta = theta[nth1+1:nth2], optimize = FALSE)
+        QQ <- Sparse(kronecker(Q1, Q2))
         idx <- which(QQ@i <= QQ@j)
         return(QQ@x[idx])
       }
@@ -438,8 +476,8 @@ setMethod(
 
       log.prior <- function(n, theta) {
         return(
-          INLA::inla.rgeneric.q(rmodel = X, cmd = "log.prior", theta = theta[1:nth1]) +
-          prior(Y, theta = theta[nth1+1:nth2])
+          rgeneric_get(X, cmd = "log_prior", theta = theta[1:nth1]) +
+          cgeneric_get(Y, cmd = "log_prior", theta = theta[nth1+1:nth2])
         )
       }
 
@@ -465,24 +503,59 @@ setMethod(
     }
 
     ### follows INLA:::inla.rgeneric.define() but no assign env
-    rmodel <- list(
-      f = list(
-        model = "rgeneric",
-        n = n,
-        rgeneric = list(
-          definition =
-            compiler::cmpfun(
-              kmodel,
-              options = list(optimize = 3L)),
-          debug = debug,
-          optimize = TRUE
+    rmodel <- structure(
+      list(
+        f = list(
+          model = "rgeneric",
+          n = n,
+          rgeneric = structure(
+            list(
+              definition =
+                compiler::cmpfun(
+                  kmodel,
+                  options = list(optimize = 3L)),
+              debug = debug,
+              optimize = TRUE
+            ),
+            # inla.rgeneric is needed to support INLA before August 2025
+            class = c("inla.rgeneric.f", "inla.rgeneric")
+          )
+        )
+      ),
+      class = c("rgeneric", "inla.rgeneric")
+    )
+
+    if((!is.null(X$f$extraconstr)) |
+       (!is.null(Y$f$extraconstr))) {
+      rmodel$f$extraconstr <-
+        kronecker_extraconstr(
+          X$f$extraconstr,
+          Y$f$extraconstr,
+          X$f$n, Y$f$n
+        )
+      if(debug) {
+        cat(nrow(rmodel$f$extraconstr),
+            " 'extraconstr' built!\n")
+      }
+    }
+
+    # Note: kron(X,Y) gives X-major ordering (the X-index varies slowly, the
+    # Y-index varies quickly), which requires the mappers to be in reverse
+    # order, multi(Y,X):
+    inlabruCheck <- packageCheck("inlabru", "2.13.0.9005")
+    if(is.na(inlabruCheck)) {
+      warning("Please install a inlabru recent version.")
+    } else {
+      rmodel$mapper <-
+      inlabru::bm_multi(
+        list(
+          inlabru::bru_get_mapper(Y),
+          inlabru::bru_get_mapper(X)
         )
       )
-    )
-    class(rmodel) <- "rgeneric"
-    class(rmodel$f$rgeneric) <- "rgeneric"
-    return(rmodel)
+    }
 
+    return(rmodel)
   }
 )
 #' @rdname kronecker
@@ -515,16 +588,15 @@ setMethod(
                        theta = NULL) {
 
       graph <- function(n, theta) {
-        g1 <- INLA::inla.rgeneric.q(X, "graph")
-        g2 <- INLA::inla.rgeneric.q(Y, "graph")
-        return(kronecker(g1, g2))
+        g1 <- rgeneric_get(X, cmd = "graph", optimize = FALSE)
+        g2 <- rgeneric_get(Y, cmd = "graph", optimize = FALSE)
+        return(Sparse(kronecker(g1, g2)))
       }
 
       Q <- function(n, theta) {
-        Q1 <- INLA::inla.rgeneric.q(rmodel = X, cmd = "Q", theta = theta[1:nth1])
-        Q2 <- INLA::inla.rgeneric.q(rmodel = Y, cmd = "Q", theta = theta[nth1+1:nth2])
-        QQ <- INLA::inla.as.sparse(
-          kronecker(Q1, Q2))
+        Q1 <- rgeneric_get(X, cmd = "Q", theta = theta[1:nth1], optimize = FALSE)
+        Q2 <- rgeneric_get(Y, cmd = "Q", theta = theta[nth1+1:nth2], optimize = FALSE)
+        QQ <- Sparse(kronecker(Q1, Q2))
         idx <- which(QQ@i <= QQ@j)
         return(QQ@x[idx])
       }
@@ -536,8 +608,8 @@ setMethod(
         return(numeric(0))
 
       log.prior <- function(n, theta) {
-        lp1 <- INLA::inla.rgeneric.q(rmodel = X, cmd = "log.prior", theta = theta[1:nth1])
-        lp2 <- INLA::inla.rgeneric.q(rmodel = Y, cmd = "log.prior", theta = theta[nth1+1:nth2])
+        lp1 <- rgeneric_get(X, cmd = "log.prior", theta = theta[1:nth1])
+        lp2 <- rgeneric_get(Y, cmd = "log.prior", theta = theta[nth1+1:nth2])
         return(lp1 + lp2)
       }
 
@@ -561,10 +633,151 @@ setMethod(
       return(ret)
     }
 
-    return(INLA::inla.rgeneric.define(
-      model = kmodel,
-      optimize = TRUE
-    ))
+    rmodel <- structure(
+      list(
+        f = list(
+          model = "rgeneric",
+          n = n,
+          rgeneric = structure(
+            list(
+              definition =
+                compiler::cmpfun(
+                  kmodel,
+                  options = list(optimize = 3L)),
+              debug = debug,
+              optimize = TRUE
+            ),
+            # inla.rgeneric is needed to support INLA before August 2025
+            class = c("inla.rgeneric.f", "inla.rgeneric")
+          )
+        )
+      ),
+      class = c("rgeneric", "inla.rgeneric")
+    )
 
+    if((!is.null(X$f$extraconstr)) |
+       (!is.null(Y$f$extraconstr))) {
+      rmodel$f$extraconstr <-
+        kronecker_extraconstr(
+          X$f$extraconstr,
+          Y$f$extraconstr,
+          X$f$n, Y$f$n
+        )
+      if(debug) {
+        cat(nrow(rmodel$f$extraconstr),
+            " 'extraconstr' built!\n")
+      }
+    }
+
+    # Note: kron(X,Y) gives X-major ordering (the X-index varies slowly, the
+    # Y-index varies quickly), which requires the mappers to be in reverse
+    # order, multi(Y,X):
+    inlabruCheck <- packageCheck("inlabru", "2.13.0.9005")
+    if(is.na(inlabruCheck)) {
+      warning("Please install a inlabru recent version.")
+    } else {
+      rmodel$mapper <-
+      inlabru::bm_multi(
+        list(
+          inlabru::bru_get_mapper(Y),
+          inlabru::bru_get_mapper(X)
+        )
+      )
+    }
+
+    return(rmodel)
   }
 )
+#' Kronecker (product) between `extraconstr`,
+#' implemented for [kronecker()] methods.
+#' @name extraconstr
+#' @param c1,c2 named list with two elements:
+#' `A` and `e`, where `nrow(A)` should be equal
+#' to `length(e)`. These are constraint definitions.
+#' @param n1,n2 integer with each model's length.
+#' @returns The constraint definition for the
+#' whole latent model built from the Kronecker product.
+#' A length two named list. 'A' a matrix and
+#' 'e' a vector where nrow(A)=length(e) and
+#' ncol(A)=(n1*n2).
+kronecker_extraconstr <- function(c1, c2, n1, n2) {
+  if(is.null(c1)) {
+    if(is.null(c2)) {
+      return(NULL)
+    } else {
+      ret <- list(
+        A = kronecker(diag(n1), c2$A),
+        e = rep(c2$e, n1)
+      )
+    }
+  } else {
+    if(is.null(c2)) {
+      ret <- list(
+        A = kronecker(c1$A, diag(n2)),
+        e = rep(c1$e, each = n2)
+      )
+    } else {
+      Aret <- rbind(
+        kronecker(c1$A, diag(ncol(c2$A))),
+        kronecker(diag(ncol(c1$A)), c2$A)
+      )
+      eret <- c(rep(c1$e, each = ncol(c2$A)),
+                rep(c2$e, ncol(c1$A)))
+      ret <- list(
+        ## remove one (the last) redundant constraint
+        A = Aret[-nrow(Aret), , drop = FALSE],
+        e = eret[-nrow(Aret)]
+      )
+    }
+  }
+  return(ret)
+}
+
+
+#' @title Combine two or more `cgeneric` or `rgeneric` models
+#' @description Constructs a multiple kronecker product model from a list of model
+#' objects. The resulting model contains a corresponding [inlabru::bm_multi()]
+#' mapper. This can be used as an alternative to a binary tree of kronecker
+#' product models.
+#' @param models A list of `cgeneric` or `rgeneric` models, optionally with names
+#' @details The last model in the list has the slowest index variation, and the
+#'   first model has the fastest index variation. This matches the latent
+#'   variable ordering of standard `INLA:f()` model components with
+#'   `(main, group, replicate)`.
+#' @param \dots Arguments passed on to every `kronecker()` call.
+#' @returns A 'cgeneric' or 'rgeneric' model object, containing a
+#'   multi-kronecker product model, with a corresponding [inlabru::bm_multi()]
+#'   mapper.
+#' @rdname multi_generic_model
+#' @export
+#' @example demo/kronecker3.R
+multi_generic_model <- function(models, ...) {
+  stopifnot(is.list(models))
+  stopifnot(length(models) >= 1L)
+
+  models <- lapply(models, function(model) {
+    if (inherits(model, c("rgeneric", "inla.rgeneric"))) {
+      rgeneric(model)
+    } else if (inherits(model, c("cgeneric", "inla.cgeneric"))) {
+      cgeneric(model)
+    } else {
+      stop("Each 'models' element must be convertible to class 'cgeneric' or 'rgeneric'")
+    }
+  })
+
+  ret <- models[[1]]
+  for (i in seq_along(models)[-1]) {
+    ret <- Matrix::kronecker(models[[i]], ret, ...)
+  }
+
+  inlabruCheck <- packageCheck("inlabru", "2.13.0.9005")
+  if(is.na(inlabruCheck)) {
+    warning("Please install a inlabru recent version.")
+  } else {
+    ret[["mapper"]] <- inlabru::bm_multi(
+      lapply(models, inlabru::bru_get_mapper)
+    )
+  }
+
+  return(ret)
+}
